from pathlib import Path
import string
from typing import List
from pyqrlew.io.postgresql import PostgreSQL
import pytest
from sqlalchemy import (
    Column,
    ForeignKey,
    MetaData,
    Table,
    create_engine,
    types,
    text
)
import pandas as pd
import numpy as np

np.random.seed(10)
LENGTH_1 = 900
LENTGH_2 = 100
LENGTH_PUBLIC = 3

@pytest.fixture
def tests_path() -> Path:
    return Path(__file__).parent

@pytest.fixture
def queries_path(tests_path) -> Path:
    """Provide the path to test queries"""
    return tests_path / 'queries' / 'base_queries.sql'

@pytest.fixture
def queries(queries_path) -> List[str]:
    with open(queries_path, 'r') as f:
        return [query for query in f if not query.startswith('--')]

@pytest.fixture(scope="session", autouse=True)
def engine():
    database = PostgreSQL()
    return database.engine()


@pytest.fixture(scope="session", autouse=True)
def metadata():
    output = MetaData()
    return output


@pytest.fixture(scope="session", autouse=True)
def simple_tables(metadata, engine):
    tab_A = Table(
        "table_A",
        metadata,
        Column("a1", types.String, primary_key=True),
        Column("a2", types.Integer, nullable=True),
        Column("a3", types.Float, nullable=False),
        Column("a4", types.String, ForeignKey("table_C.c1", ondelete='CASCADE'), nullable=False),
    )
    tab_B = Table(
        "table_B",
        metadata,
        Column("b1", types.String, primary_key=True),
        Column("b2", types.String, ForeignKey("table_A.a1", ondelete='CASCADE'), nullable=True),
    )
    tab_D = Table(  # table to ignore
        "table_D",
        metadata,
        Column("d1", types.String, primary_key=True),
    )
    tab_C = Table(
        "table_C",
        metadata,
        Column("c1", types.String, primary_key=True),
        Column(
            "c2",
            types.String,
            ForeignKey("table_D.d1", ondelete='CASCADE'),
            nullable=True,
        ),
    )
    tab_E = Table(
        "table_E",
        metadata,
        Column("e1", types.Integer, primary_key=True),
    )

    tab_Abis = Table(
        "table_Abis",
        metadata,
        Column("a4", types.String, nullable=False),
        Column("a1", types.String),
        Column("a2", types.Integer, nullable=True),
        Column("a3", types.Float, nullable=False),
    )

    with engine.begin() as conn:
        conn.execute(text('DROP TABLE IF EXISTS "table_A", "table_B", "table_D", "table_C", "table_E", "table_Abis" CASCADE'))

    metadata.create_all(engine)
    df_d = pd.DataFrame(
        {
            "d1": ["Paris", "Rome", "Madrid"],
        }
    )
    df_d.to_sql(name="table_D", con=engine, if_exists="append", index=False)
    df_c = pd.DataFrame(
        {
            "c1": ["France", "Italy", "Spain"],
            "c2": ["Paris", "Rome", "Madrid"],
        }
    )
    df_c.to_sql(name="table_C", con=engine, if_exists="append", index=False)
    df_A = pd.DataFrame(
        {
            "a1": ["User1", "User2", "User3"],
            "a2": [3, 5, 9],
            "a3": [6.0, 10.0, 8.0],
            "a4": ["Spain", "Italy", "Italy"],
        }
    )
    df_A.to_sql(name="table_A", con=engine, if_exists="append", index=False)
    df_Abis = pd.DataFrame(
        {
            "a4": ["Spain", "Italy", "Italy", "Spain", "Spain"],
            "a1": ["User1", "User1", "User1", "User2", "User2"],
            "a2": [3, 5, 9, None, 10],
            "a3": [16.0, 110.0, 88.0, 99.0, 231.0],
        }
    )
    df_Abis.to_sql(
        name="table_Abis", con=engine, if_exists="append", index=False
    )
    df_B = pd.DataFrame(
        {
            "b1": ["Transaction1", "Transaction2"],
            "b2": ["User2", "User1"],
        }
    )
    df_B.to_sql(name="table_B", con=engine, if_exists="append", index=False)
    df_E = pd.DataFrame(
        {
            "e1": list(range(99)),
        }
    )
    df_E.to_sql(name="table_E", con=engine, if_exists="append", index=False)
    return tab_A, tab_B, tab_C, tab_D, tab_E, tab_Abis

@pytest.fixture(scope="session", autouse=True)
def tables(metadata, engine):

    _ = Table(
        "primary_public_table",
        metadata,
        Column("id", types.Integer(), primary_key=True),
        Column("text", types.String(), nullable=True),
    )
    _ = Table(
        "secondary_public_table",
        metadata,
        Column("id", types.Integer(), primary_key=True),
        Column("text", types.String(), nullable=True),
    )
    _ = Table(
        "primary_table",
        metadata,
        Column("id", types.Integer(), primary_key=True, nullable=True),
        Column("integer", types.Integer(), nullable=False),
        Column("float", types.Float(), nullable=False),
        Column("datetime", types.DateTime(), nullable=False),
        Column("date", types.Date(), nullable=False),
        Column("boolean", types.Boolean(), nullable=False),
        Column("text", types.String(), nullable=False),
        Column(
            "public_fk",
            types.Integer(),
            ForeignKey("primary_public_table.id", ondelete='CASCADE'),
            nullable=False,
        ),
    )
    _ = Table(
        "secondary_table",
        metadata,
        Column("id", types.Integer(), ForeignKey("primary_table.id", ondelete='CASCADE'), nullable=True),
        Column("integer", types.Integer(), nullable=True),
        Column("float", types.Float(), nullable=True),
        Column("datetime", types.DateTime(), nullable=True),
        Column("date", types.Date(), nullable=True),
        Column("boolean", types.Boolean(), nullable=True),
        Column("text", types.String(), nullable=True),
        Column(
            "public_fk",
            types.Integer(),
            ForeignKey("secondary_public_table.id", ondelete='CASCADE'),
            nullable=True,
        ),
    )

    with engine.begin() as conn:
        conn.execute(text("DROP TABLE IF EXISTS primary_table, secondary_table, primary_public_table, secondary_public_table CASCADE"))

    metadata.create_all(engine)
    private, public = create_df(sql=True)
    public.to_sql(
        name="primary_public_table",
        con=engine,
        if_exists="append",
        index=False,
    )
    public.to_sql(
        name="secondary_public_table",
        con=engine,
        if_exists="append",
        index=False,
    )
    private.to_sql(
        name="primary_table", con=engine, if_exists="append", index=False
    )
    private.to_sql(
        name="secondary_table", con=engine, if_exists="append", index=False
    )

def create_df(sql: bool) -> pd.DataFrame:
    public = pd.DataFrame(
        {
            "id": np.linspace(1, LENGTH_PUBLIC, LENGTH_PUBLIC),
            "text": [
                "".join(
                    np.random.choice(
                        list(string.ascii_uppercase + string.digits), size=8
                    ).tolist()
                )
                for _ in range(LENGTH_PUBLIC)
            ],
        }
    )

    private = pd.DataFrame(
        {
            "id": np.linspace(1, LENGTH_1, LENGTH_1, dtype=int),
            "integer": np.random.randint(low=0, high=100, size=LENGTH_1),
            "float": np.random.random(size=LENGTH_1),
            "datetime": np.random.choice(
                pd.date_range("1980-01-01", "1980-02-01", freq="s"),
                LENGTH_1,
                replace=True,
            ),
            "text": [
                "".join(
                    np.random.choice(
                        list(string.ascii_uppercase + string.digits), size=8
                    ).tolist()
                )
                for _ in range(LENGTH_1)
            ],
            "date": np.random.choice(
                pd.date_range("1980-01-01", "2000-01-01", freq="D"),
                LENGTH_1,
                replace=True,
            ),
            "boolean": np.random.choice(
                [True, False], LENGTH_1, replace=True, p=[0.7, 0.3]
            ),
            "public_fk": np.random.choice(
                public["id"].values, LENGTH_1, replace=True
            ),
        }
    )

    if sql:
        return private, public

    private = private.drop("public_fk", axis=1)
    private["date"] = private["date"].astype(str)
    return private


"""Module with the definition of some rust objects that are not exposed to
python. This is only for documentation purposes."""
import typing as t

PrivacyUnit = t.Union[
    t.Sequence[t.Tuple[str, t.Sequence[t.Tuple[str, str, str]], str]],
    t.Tuple[t.Sequence[t.Tuple[str, t.Sequence[t.Tuple[str, str, str]], str]], bool],
    t.Tuple[t.Sequence[t.Tuple[str, t.Sequence[t.Tuple[str, str, str]], str, str]], bool]
]
SyntheticData = t.Sequence[t.Tuple[t.Sequence[str], t.Sequence[str]]]
DPEvent = t.Mapping[str, t.Union[str, float, t.Sequence['DPEvent']]]

class DpEvent(t.Protocol):
    """Internal object containing a description of differentially private
    mechanisms such as Laplace and Gaussian, etc, and their composition.
    This is compatible with dp-accounting, part of the Google differential
    privacy library, containing tools for tracking differential privacy
    budgets.
    """

    def to_dict(self) -> DPEvent:
        """Returns a Dict representation of DP mechanisms."""
        ...

    def to_named_tuple(self) -> t.NamedTuple:
        """Returns NamedTuple of DP mechanisms compatible with `dp-accounting`
        """
        ...



from pyqrlew.io.postgresql import PostgreSQL
from pyqrlew.wrappers import Relation
import pytest


@pytest.fixture
def extract_dataset():
    database = PostgreSQL()
    dataset = database.extract()
    return dataset


def test_wrong_query(extract_dataset):
    query = "SELECT age, COUNT(*) FROM not_existing_table GROUP BY age"
    with pytest.raises(RuntimeError) as e_info:
        _ = Relation.from_query(query, extract_dataset)

    query = "wrong query"
    with pytest.raises(RuntimeError) as e_info:
        _ = Relation.from_query(query, extract_dataset)

    # this panics, issue: https://github.com/Qrlew/qrlew/issues/287
    # query = "SELECT not_existing_col, COUNT(*) FROM  extract.census"
    # with pytest.raises(RuntimeError) as e_info:
    #     rel = Relation.from_query(query, extract_dataset)
    #     print(e_info)





from pyqrlew.utils import tables_prefix
from pyqrlew import Dialect

def test_tables_prefix():
    query_str = """
    WITH mytab AS (SELECT * FROM A.b.c),
    mytab2 AS (SELECT * FROM (SELECT * FROM d.e.f) AS subtab )
    SELECT * FROM mytab JOIN mytab2 USING(id)
    """
    tables = tables_prefix(query_str, Dialect.PostgreSql)
    assert tables == ["A","d"]



import tempfile
import os
import platform
import subprocess
import typing as t
from .pyqrlew import Dialect

MAGENTA_COLOR = '\033[35m'
BLUE_COLOR = '\033[34m'
RESET_COLOR = '\033[0m'


def print_query(query: str) -> None:
    """
    Print a formatted and syntax-highlighted SQL query.

    This function takes an SQL query as input and prints it to the console with syntax
    highlighting applied to certain keywords for better readability.

    Parameters
    ----------
    query: str
        The SQL query to be printed and highlighted.

    Returns
    -------
    None
        This function does not return any value. It directly prints the formatted
        and highlighted SQL query to the console.
    """
    keywords = ["SELECT", "AS", "GROUP BY", "LIMIT", "ORDER BY", "WHERE"]
    colored_query = query
    colored_query = colored_query.replace("WITH", "WITH\n ")
    colored_query = colored_query.replace(" SELECT", "\nSELECT")
    colored_query = colored_query.replace("),", "),\n ")
    for word in keywords:
        colored_query = colored_query.replace(word, MAGENTA_COLOR + word + RESET_COLOR)
    colored_query = colored_query.replace("WITH", BLUE_COLOR + "WITH" + RESET_COLOR)
    print(colored_query)


def display_graph(dot_string):
    # This is an optional dependency
    from graphviz import Source  #type: ignore

    # Create a temporary file for the SVG
    with tempfile.NamedTemporaryFile(delete=False, suffix='.svg') as temp:
        # Create a Source object from the DOT string
        src = Source(dot_string)
        
        # Render the source to a temporary SVG file
        svg_path = src.render(filename=temp.name, format='svg', cleanup=False)

    # Create a temporary HTML file to embed the SVG
    html_path = f"{temp.name}.html"
    with open(html_path, 'w') as html_file:
        # Use absolute path for SVG in the HTML to ensure it can be loaded
        html_content = f"""
        <html>
            <head>
                <style>
                    body, html {{
                        height: 100%;
                        margin: 0;
                        display: flex;
                        justify-content: center;
                        align-items: center;
                    }}
                </style>
            </head>
            <body>
                <img src="file://{svg_path}" alt="Graph" style="max-width: 100%; max-height: 100%;"/>
            </body>
        </html>
        """
        html_file.write(html_content)
    
    # Open the rendered HTML file in the default system browser
    try:
        if platform.system() == 'Darwin':       # macOS
            subprocess.run(['open', html_path])
        elif platform.system() == 'Windows':   # Windows
            os.startfile(html_path)
        else:                                   # Linux variants
            subprocess.run(['xdg-open', html_path])
    finally:
        pass

def tables_prefix(query: str, dialect: Dialect) -> t.List[str]:
    """Util to extract the prefix of fully qualified table names from a query.
    Only base tables are considered. If table names are quoted, the quoting
    must be coherent with the dialect.

    Args:
        query (str):
        dialect (Dialect): 

    Returns:
        t.List[str]: prefix of table names
    """
    from pyqrlew.pyqrlew import tables_prefix
    return tables_prefix(query, dialect)


import shutil
from pathlib import Path
from sqlalchemy import select, column, table, text
from qrlew_datasets.files import SQL  # type: ignore  # module is installed, but missing library stubs or py.typed marker 
from qrlew_datasets.databases import PostgreSQL as EmptyPostgreSQL  # type: ignore  # module is installed, but missing library stubs or py.typed marker 
from pyqrlew import dataset_from_database, Dataset

NAME: str = 'pyqrlew-db'
USER: str = 'postgres'
PASSWORD: str = 'pyqrlew-db'
PORT: int = 5433

POSSIBLE_VALUES_THRESHOLD: int = 20

class PostgreSQL(EmptyPostgreSQL):
    """
    A class representing a PostgreSQL database with PyQrlew integration.

    This class provides methods for interacting with an PostgreSQL database
    using the PyQrlew library.

    Example
    ----------
        >>> import pyqrlew as qrl
        >>> from pyqrlew.io import PostgreSQL
        >>> import pandas as pd
        >>> DB = PostgreSQL() # Setup a default database
        >>> DB.load_extract() # Insert a demo dataset
        >>> dataset = Dataset.from_database(name='extract', engine=DB.engine(), schema_name='extract')
        >>> dataset.relations()
        [(['extract', 'extract', 'beacon'], <Relation at 0x784b78176fd0>),
        (['extract', 'extract', 'census'], <Relation at 0x784b78177490>)]
        >>> dataset.extract.census.relation().schema()
        {age: int, workclass: str, fnlwgt: str, education: str, education_num: int, marital_status: str, occupation: str, relationship: str, race: str, sex: str, capital_gain: int, capital_loss: int, hours_per_week: int, native_country: str, income: str}
        >>> query = "SELECT AVG(age) AS avg_age FROM extract.census"
        >>> pd.read_sql(query, DB.engine())
        	avg_age
        0	48.005155
    """
    def __init__(self, name:str=NAME, user:str=USER, password:str=PASSWORD, port:int=PORT) -> None:
        """Initialize an SQLite instance.

        Args:
            name (str, optional): The name of the instance. Defaults to NAME.
            user (str, optional): user name in the database. Defaults to USER.
            password (str, optional): password used to connect to the service.
                Defaults to PASSWORD.
            port (int, optional): port number at which the database listens.
                Defaults to PORT.
        """
        super().__init__(name, user, password, port)

    def load_resource(self, schema: str, src: Path) -> 'PostgreSQL':
        """It loads sql data resource into the specified schema of the
        PostgreSQL database. It checks if the schema exists, and if not, copies
        the resource to a temporary location and then loads it into
        the database.

        Args:
            schema (str): The name of the database schema where
                the resource is to be loaded.
            src (Path): The file path of the resource to be
                loaded (an sql script) into the database.

        Returns:
            PostgreSQL: The instance of the PostgreSQL class,
            allowing for method chaining.
        """
        with self.engine().connect() as conn:
            res = list(
                conn.execute(
                    select(column('schema_name'))
                    .select_from(table('schemata', schema='information_schema'))
                    .where(column('schema_name') == schema))
            )
            schema_exists = len(res)==1
        if not schema_exists:
            dst = Path('/tmp') / schema
            shutil.copyfile(src, dst)
            self.load(dst)
        return self

    def load_extract(self) -> 'PostgreSQL':
        """It loads the 'extract' dataset in the 'extract' schema.
        It internally calls the load_resource method 

        Returns:
            PostgreSQL: The instance of the PostgreSQL class,
            allowing for method chaining.
        """
        return self.load_resource('extract', SQL('extract').local())

    def load_financial(self) -> 'PostgreSQL':
        """It loads the 'financial' dataset in the 'financial' schema.
        It internally calls the load_resource method 

        Returns:
            PostgreSQL: The instance of the PostgreSQL class,
            allowing for method chaining.
        """
        return self.load_resource('financial', SQL('financial').local())

    def load_hepatitis(self) -> 'PostgreSQL':
        """It loads the 'hepatitis' dataset in the 'hepatitis_std' schema.
        It internally calls the load_resource method 

        Returns:
            PostgreSQL: The instance of the PostgreSQL class,
            allowing for method chaining.
        """
        return self.load_resource('hepatitis_std', SQL('hepatitis').local())

    def load_imdb(self) -> 'PostgreSQL':
        """It loads the 'imdb' dataset in the 'imdb_ijs' schema.
        It internally calls the load_resource method 

        Returns:
            PostgreSQL: The instance of the PostgreSQL class,
            allowing for method chaining.
        """
        return self.load_resource('imdb_ijs', SQL('imdb').local())

    def load_retail(self) -> 'PostgreSQL':
        """It loads the 'retail' dataset in the 'retail' schema.
        It internally calls the load_resource method 

        Returns:
            PostgreSQL: The instance of the PostgreSQL class,
            allowing for method chaining.
        """
        return self.load_resource('retail', SQL('retail').local())

    def extract(self) -> Dataset:
        """It loads the extract dataset in the database and it returns a qrlew
        Dataset with data ranges and categorical possible values if the unique
        column values are less than POSSIBLE_VALUES_THRESHOLD.
        """
        self.load_extract()
        return dataset_from_database('extract', self.engine(), 'extract', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def financial(self) -> Dataset:
        """It loads the financial dataset in the database and it returns a qrlew
        Dataset with data ranges and categorical possible values if the unique
        column values are less than POSSIBLE_VALUES_THRESHOLD.
        """
        self.load_financial()
        return dataset_from_database('financial', self.engine(), 'financial', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def hepatitis(self) -> Dataset:
        """It loads the hepatitis dataset in the database and it returns a qrlew
        Dataset with data ranges and categorical possible values if the unique
        column values are less than POSSIBLE_VALUES_THRESHOLD.
        """
        self.load_hepatitis()
        return dataset_from_database('hepatitis_std', self.engine(), 'hepatitis_std', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def imdb(self) -> Dataset:
        """It loads the imdb dataset in the database and it returns a qrlew
        Dataset with data ranges and categorical possible values if the unique
        column values are less than POSSIBLE_VALUES_THRESHOLD.
        """
        self.load_imdb()
        return dataset_from_database('imdb_ijs', self.engine(), 'imdb_ijs', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)

    def retail(self) -> Dataset:
        """It loads the retail dataset in the database and it returns a qrlew
        Dataset with data ranges and categorical possible values if the unique
        column values are less than POSSIBLE_VALUES_THRESHOLD.
        """
        self.load_retail()
        return dataset_from_database('retail', self.engine(), 'retail', ranges=True, possible_values_threshold=POSSIBLE_VALUES_THRESHOLD)
